##############################
# Laplace Distribution
##############################

# truncated (and rounded) Laplace law
# LP[i] is an integer 
def Laplace(mu, sigma, i):
    return exp(-abs(i - mu) / sigma)

def Laplace_precomp(mu, sigma, i_min, i_max):
    LP = {}
    CLP = {}
    for i in range(i_min, i_max + 1):
        LP[i] = Laplace(mu, sigma, i)
    MF = 1000 / min(LP[i] for i in LP) # multiplicative factor
    for i in LP:
        LP[i] = int(round(MF * LP[i]))
    CLP[i_min] = LP[i_min]
    for i in range(i_min, i_max):
        CLP[i + 1] = CLP[i] + LP[i + 1]
    return LP, CLP

##########################################
# Target Distribution for |eV| (q_1)
##########################################

# i_min, i_max <-- q1_target_range()
# where q1(i)<1/2^128 for i not in [i_min,i_max[
def q1_target_range(n2, w):
    i_avg = int(2 * n2 / 3)
    i = i_avg
    j = 0
    while True:
        l = int((i + j) / 2)
        if q1_target(l, n2, w) < 2^(-128):
            j = l
        else:
            i = l
        if i - j <= 1:
            break
    i_min = l
    i = n2
    j = i_avg
    while True:
        l = int((i + j) / 2)
        if q1_target(l, n2, w) < 2^(-128):
            i = l
        else:
            j = l
        if i - j <= 1:
            break
    i_max = i
    return i_min + 1, i_max
def q1_target_term(i, j, n2, w):
    return binomial(n2, j) * binomial(j, i) * binomial(i, 2 * j - w) * 8.0^j
def q1_target(i, n2, w):
    j_min = max(i, ceil(w / 2))
    j_max = min(n2, floor((w + i) / 2))
    n = n2 * 2
    return sum(q1_target_term(i, j, n2, w) for j in range(j_min, j_max + 1)) / (binomial(n, w) * 4.0^w)
def q1_target_precomp(n2, w, i_min, i_max):
    QTV = {}
    for i in range(i_min, i_max):
        QTV[i] = q1_target(i, n2, w)
    return QTV

##########################################
# Emulated Distribution for |eV|
##########################################

def q1_emul_term_precomp(n2, kV, w):
    T = {}
    for i in range(n2 - kV + 1):
        T[i] = binomial(n2 - kV, i) * 2.0^i
    return T
def q1_emul(i, QETV, LPV, n2, kV, w):
    t_min = max(0, i - n2 + kV)
    t_max = min(i, kV)
    return sum(QETV[i - t] * LPV[t] for t in range(t_min, t_max + 1)) / 3.0^(n2 - kV)
def q1_emul_precomp(QETV, LPV, n2, kV, w, i_min, i_max):
    QEV = {}
    NF = 0 # normalization factor
    for i in range(i_min, i_max):
        QEV[i] = q1_emul(i, QETV, LPV, n2, kV, w)
        NF += QEV[i]
    for i in range(i_min, i_max):
        QEV[i] /= NF
    return QEV

##########################################
# Precompute everything for V
##########################################

def precomp_V(sigmaV, alpha, n2, kV, w):
    i_min, i_max = q1_target_range(n2, w)
    QTV = q1_target_precomp(n2, w, i_min, i_max)
    muV = (1 - alpha) * kV
    LPV, CLPV = Laplace_precomp(muV, sigmaV, 0, kV)
    QETV = q1_emul_term_precomp(n2, kV, w)
    QEV = q1_emul_precomp(QETV, LPV, n2, kV, w, i_min, i_max)
    rV = {}
    for i in QEV:
        rV[i] = QTV[i] / QEV[i]
    MV = max(rV[i] for i in rV)
    for i in rV:
        rV[i] /= MV
    return rV, CLPV, QEV, QTV

##########################################
# Target Distribution for lUV(eU,eV) (q_2)
##########################################

def q2_target_term(j, t, n2, w):
    return binomial(t, 2*j - w) * binomial(n2 - t, j - t) * 8.0^j
def q2_target_precomp(t, n2, w):
    j_min = max(t, ceil(w / 2))
    j_max = min(n2, floor((w + t) / 2))
    D = {}
    NF = 0
    for j in range(j_min, j_max + 1):
        D[2 * j - w] = q2_target_term(j, t, n2, w)
        NF += D[2 * j - w]
    for j in range(j_min, j_max + 1):
        D[2 * j - w] /= NF
    for j in range(j_min, j_max + 1):
        if D[2 * j - w] < 2^(-128):
            D.pop(2 * j - w)
    return D
def q2_target_precomp_all(t_min, t_max, n2, w):
    QTU = {}
    for t in range(t_min, t_max):
        QTU[t] = q2_target_precomp(t, n2, w)
    return QTU

##########################################
# Emulated Distribution for lUV(eU,eV)
##########################################

def q2_emul_term(i, j, n2, kU, w):
    return binomial(i, 2 *j - w) * binomial(n2 - kU - i, j - kU - i) * 8.0^j
def q2_emul_term_precomp(i, n2, kU, w):
    D = {}
    j_min = max(kU + i, ceil(w / 2))
    j_max = min(n2, floor((w + i) / 2))
    NF = 0
    for j in range(j_min, j_max + 1):
        D[2 * j - w] = q2_emul_term(i, j, n2, kU, w)
        NF += D[2 * j - w]
    for j in range(j_min, j_max + 1):
        D[2 * j - w] /= NF
    return D
def q2_emul_term_precomp_all(t_min, t_max, n2, kU, w):
    QETU = {}
    for t in range(t_max):
        QETU[t] = q2_emul_term_precomp(t, n2, kU, w)
    return QETU
def q2_emul(s, t, QETU, LPU, n2, kU, w):
    x = 0
    if mod(w + s, 2) == 0:
        for i in range(t + 1):
            if (s in QETU[t - i]) and (i in LPU):
                x += QETU[t - i][s] * LPU[i]
    return x
def q2_emul_precomp(t, QETU, LPU, n2, kU, w):
    D = {}
    NF = 0
    for s in range(mod(w, 2), t + 1, 2):
        x = q2_emul(s, t, QETU, LPU, n2, kU, w)
        if x != 0:
            NF += x
            D[s] = x
    for s in D:
        D[s] /= NF
    return D
# Depending on the Sagemath version, the library function find_root
# fails unexpectedly for some inputs. We use an ad-hoc version instead
# which impose condition on the input which are always met here
# assumes a < b, f(a) > 0, f(b) < 0, and a single root in [a,b]
def find_rootU(f, a, b, epsilon):
    fa = f(a)
    fb = f(b)
    c = float((a + b) / 2)
    fc = f(c)
    while (b - a) > epsilon:
        if fc > 0:
            a = c
            fa = fc
        else:
            b = c
            fb = fc
        c = float((a + b) / 2)
        fc = f(c)
    return c
# def muU(t, n, w):
#     print t, n, w
#     s = var('s')
#     return max(0, t - 3 / 2 * find_root(8*(-t+s)*(-t+s-1)*(n-w-s+1)==(s+1)*s*(w+s+1-2*t),0,n - w))
def muU(t, n, w):
    def f(s): return 8*(-t+s)*(-t+s-1)*(n-w-s+1)-(s+1)*s*(w+s+1-2*t)
    y = find_rootU(f, 0, n - w, 1e-6)
    return max(0, t - 3 / 2 * y)

def q2_laplace_precomp_all(sigmaU, epsilon, t_min, t_max, n2, kU, w):
    LPU = {}
    CLPU = {}
    for t in range(t_min, t_max):
        LPavg = muU(t, n2 * 2, w) + epsilon
        LPU[t], CLPU[t] = Laplace_precomp(LPavg, sigmaU, max(0, kU - n2 + t), t)
    return LPU, CLPU

##########################################
# Precompute everything for U
##########################################

def precomp_U(sigmaU, epsilon, n2, kU, w, t_min, t_max):
    QTU = q2_target_precomp_all(t_min, t_max, n2, w)
    QETU = q2_emul_term_precomp_all(t_min, t_max, n2, kU, w)
    LPU, CLPU = q2_laplace_precomp_all(sigmaU, epsilon, t_min, t_max, n2, kU, w)
    QEU = {}
    rU = {}
    for t in range(t_min, t_max):
        QEU[t] = q2_emul_precomp(t, QETU, LPU[t], n2, kU, w)
        rU[t] = {}
        for i in QTU[t]:
            rU[t][i] = QTU[t][i] / QEU[t][i]
        MU = max(rU[t][i] for i in rU[t])
        for i in rU[t]:
            rU[t][i] /= MU
    return rU, CLPU, QEU, QTU
