import random

# parameters and key generation

class coeffUV():
    def __init__(self, F, strict):
        if strict:
            self.a = 1
            self.c = 1
            self.b = 0
            self.d = 1
            self.coeff = 1
        else:
            self.a = random.choice(F.list_of_elements_of_multiplicative_group())
            self.c = random.choice(F.list_of_elements_of_multiplicative_group())
            self.b = random.choice(F.list())
            self.d = (F.one() + self.b * self.c) / self.a
            self.coeff = Set(F.list()).difference([-self.b / self.a, -self.d / self.c])[0]

class SecretKeyUV():
    def __init__(self, F, n2, kU, kV, id_str, strict):
        random.seed(id_str)
        self.seedV = random.randint(0,1000000)
        self.HVt = wave_random_matrix(F, n2, n2 - kV, self.seedV)
        self.seedU = random.randint(0,1000000)
        self.HUt = wave_random_matrix(F, n2, n2 - kU, self.seedU)
        self.perm = range(2 * n2)
        random.shuffle(self.perm) # in place
        self.id_str = id_str
        self.pickRandomCoeffs(n2, F, strict)

    def pickRandomCoeffs(self, n2, F, strict):
        self.coeff = []
        for i in range(n2):
            self.coeff += [coeffUV(F, strict)]

    def fullPCM_tr(self):
        n2 = self.HUt.nrows()
        Hu = self.HUt.stack(self.HUt)
        Hv = self.HVt.stack(self.HVt)
        for j in range(n2):
            Hu.rescale_row(j, self.coeff[j].d)
            Hu.rescale_row(j + n2, -self.coeff[j].b)
            Hv.rescale_row(j, -self.coeff[j].c)
            Hv.rescale_row(j + n2, self.coeff[j].a)
        return Hu.augment(Hv)

    def phiUV(self, eU, eV):
        n2 = len(eU)
        x_left = [self.coeff[i].a * eU[i] + self.coeff[i].b * eV[i] for i in range(n2)]
        x_right = [self.coeff[i].c * eU[i] + self.coeff[i].d * eV[i] for i in range(n2)]
        return vector(x_left + x_right)

    def publicKey(self):
        Ht = self.fullPCM_tr()
        Ht_syst, self.perm = echelon_form_tr(Ht, self.perm)
        self.Sti = Ht.matrix_from_rows(self.perm[:Ht.ncols()])
        return Ht_syst.matrix_from_rows(self.perm)

# Input is the transpose of H
# support is a permuted list of the code support [0,1,...,n-1]
# pivots are chosen in the order given by support
# support0 is a permuted list starting with the effective pivots positions
# H0 spans the same code as H
def echelon_form_tr(Ht, support):
    n = Ht.nrows()
    r = Ht.ncols()
    H0 = Ht.matrix_from_rows(support).transpose()
    H0.echelonize()
    pv = H0.pivots()
    npv = H0.nonpivots()
    Ht0 = H0.transpose()
    support_inv = [0] * n
    for i in range(n):
        support_inv[support[i]] = i
    Ht0 = Ht0.matrix_from_rows(support_inv)
    support0 = [support[pv[i]] for i in range(len(pv))] + [support[npv[i]] for i in range(n - len(pv))]
    return Ht0, support0

def keygen(F, n2, kU, kV, *args):
    strict = False
    if args:
        id_str = args[0]
        if len(args) >= 2:
            if args[1]:
                strict = True
    else:
        id_str = str(random.randint(0,999))
    sk = SecretKeyUV(F, n2, kU, kV, id_str, strict)
    # Call below required to complete the secret key generation
    pk = sk.publicKey()
    return pk, sk

